#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <GL/glew.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include "app.h"
#include "machine.h"

static bool init();
static void cleanup();
static void display();
static void idle();
static void draw_gears();
static void reshape(int x, int y);
static void keyb(unsigned char key, int x, int y);
static void mouse(int bn, int st, int x, int y);
static void motion(int x, int y);
static void passive_motion(int x, int y);
static Gear *pick_gear(int x, int y);

static int win_width, win_height;

static float cam_dist = 0.5;
static float cam_theta, cam_phi;
static int prev_mx, prev_my;
static bool bnstate[8];

static unsigned int start_time, prev_msec;
static Machine *machine;
static Gear *hover_gear, *sel_gear;
static HitPoint pick_hit;
static Vec3 sel_hit_pos;

int main(int argc, char **argv)
{
	glutInitWindowSize(1024, 768);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_MULTISAMPLE);
	glutCreateWindow("Antikythera");

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passive_motion);

	if(!init()) {
		return 1;
	}
	atexit(cleanup);

	glutMainLoop();
	return 0;
}

static bool init()
{
	glewInit();

	glEnable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);

	Mesh::use_custom_sdr_attr = false;

	machine = new Machine;

	const float pitch = 10.0f;

	Gear *gear1 = new Gear;
	gear1->pos = Vec3(-50, 0, 0);
	gear1->set_teeth(16, pitch);
	gear1->gen_mesh();
	machine->add_gear(gear1);

	Gear *gear2 = new Gear;
	gear2->set_teeth(32, pitch);
	gear2->pos = gear1->pos + Vec3(gear1->radius + gear2->radius - gear1->teeth_length * 0.75, 0, 0);
	gear2->thickness = 5;
	gear2->gen_mesh();
	machine->add_gear(gear2);

	Gear *gear3 = new Gear;
	gear3->set_teeth(8, pitch);
	gear3->pos = gear2->pos + Vec3(0, gear2->radius + gear3->radius - gear2->teeth_length * 0.75, 0);
	gear3->gen_mesh();
	machine->add_gear(gear3);

	Gear *subgear = new Gear;
	subgear->set_teeth(10, pitch);
	subgear->pos = Vec3(0, 0, (gear2->thickness + subgear->thickness) / 2 + 1);
	subgear->gen_mesh();
	gear2->attach(subgear);
	machine->add_gear(subgear);

	machine->add_motor(0, 1.0);

	start_time = glutGet(GLUT_ELAPSED_TIME);
	return true;
}

static void cleanup()
{
	delete machine;
}

static void update(float dt)
{
	machine->update(dt);

	if(sel_gear) {
	}

	hover_gear = pick_gear(prev_mx, prev_my);
}

static void set_light(int idx, const Vec3 &pos, const Vec3 &color)
{
	unsigned int lt = GL_LIGHT0 + idx;
	float posv[] = { pos.x, pos.y, pos.z, 1 };
	float colv[] = { color.x, color.y, color.z, 1 };

	glEnable(lt);
	glLightfv(lt, GL_POSITION, posv);
	glLightfv(lt, GL_DIFFUSE, colv);
	glLightfv(lt, GL_SPECULAR, colv);
}

static void display()
{
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME) - start_time;
	float dt = (float)(msec - prev_msec) / 1000.0f;
	prev_msec = msec;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	set_light(0, Vec3(-50, 75, 100), Vec3(1.0, 0.8, 0.7) * 0.8);
	set_light(1, Vec3(100, 0, 30), Vec3(0.6, 0.7, 1.0) * 0.6);
	set_light(2, Vec3(-10, -10, 60), Vec3(0.8, 1.0, 0.8) * 0.3);

	update(dt);

	draw_gears();

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

static void idle()
{
	glutPostRedisplay();
}

static void draw_gears()
{
	/* world scale is in meters, gears are in millimeters, scale by 1/1000 */
	glPushMatrix();
	glScalef(0.001, 0.001, 0.001);

	if(sel_gear || hover_gear) {
		glPushAttrib(GL_ENABLE_BIT | GL_LINE_BIT | GL_POLYGON_BIT);

		glDisable(GL_LIGHTING);
		glFrontFace(GL_CW);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glLineWidth(3.0);

		if(sel_gear) {
			glColor3f(0.2, 1.0, 0.3);
			sel_gear->draw();
		} else {
			glColor3f(1.0, 0.75, 0.2);
			hover_gear->draw();
		}

		glPopAttrib();
	}

	machine->draw();

	glPopMatrix();
}

static void reshape(int x, int y)
{
	win_width = x;
	win_height = y;

	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.01, 100.0);
}

static void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 'w':
		opt_gear_wireframe = !opt_gear_wireframe;
		glutPostRedisplay();
		break;
	}
}

static void mouse(int bn, int st, int x, int y)
{
	int bidx = bn - GLUT_LEFT_BUTTON;
	bool down = st == GLUT_DOWN;

	prev_mx = x;
	prev_my = y;
	bnstate[bidx] = down;

	if(bidx == 0) {
		if(down) {
			sel_gear = pick_gear(x, y);
			sel_hit_pos = pick_hit.pos;
		} else {
			sel_gear = 0;
		}
	}

	if(bidx == 3 || bidx == 4) {	/* wheel */
		if(hover_gear) {
			float dz = bidx == 4 ? 1 : -1;
			hover_gear->set_position(hover_gear->get_position() + hover_gear->get_axis() * dz);
			machine->invalidate_meshing();
		}
	}
}

static void motion(int x, int y)
{
	int dx = x - prev_mx;
	int dy = y - prev_my;
	prev_mx = x;
	prev_my = y;

	if(!dx && !dy) return;

	if(sel_gear) {
		float speed = 0.5;
		Vec3 offs = Vec3(dx * speed, -dy * speed, 0.0);
		offs = sel_gear->get_dir_matrix() * offs;

		sel_gear->set_position(sel_gear->get_position() + offs);
		machine->invalidate_meshing();

	} else {
		if(bnstate[0]) {
			cam_theta += dx * 0.5;
			cam_phi += dy * 0.5;

			if(cam_phi < -90) cam_phi = -90;
			if(cam_phi > 90) cam_phi = 90;
			glutPostRedisplay();
		}
		if(bnstate[2]) {
			cam_dist += dy * 0.01;
			if(cam_dist < 0.0) cam_dist = 0.0;
			glutPostRedisplay();
		}
	}
}

static void passive_motion(int x, int y)
{
	prev_mx = x;
	prev_my = y;
}

static Gear *pick_gear(int x, int y)
{
	double pt[3];
	double viewmat[16], projmat[16];
	int vp[4];
	Ray ray;

	y = win_height - y;

	glGetDoublev(GL_MODELVIEW_MATRIX, viewmat);
	glGetDoublev(GL_PROJECTION_MATRIX, projmat);
	glGetIntegerv(GL_VIEWPORT, vp);

	gluUnProject(x, y, 0, viewmat, projmat, vp, pt, pt + 1, pt + 2);
	ray.origin = Vec3(pt[0], pt[1], pt[2]) * 1000.0f;

	gluUnProject(x, y, 1, viewmat, projmat, vp, pt, pt + 1, pt + 2);
	ray.dir = Vec3(pt[0], pt[1], pt[2]) * 1000.0f - ray.origin;

	return machine->intersect_gear(ray, &pick_hit);
}
